//******************************************************************************
/*!
	@file		mms101.c
	@brief		MMS101 Sample Source File

				copyright: MITSUMI Electric Co.,LTD
	@attention	none
	@warning	MITSUMI CONFIDENTIAL
*/
//******************************************************************************
/*******************************************************************************
* History : DD.MM.YYYY Version  Description
*         : 07.10.2021 1.0.0.0  First Release
*******************************************************************************/
#define __MMS101_C__

//==============================================================================
//{### MODULES USED
//==============================================================================
#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <string.h>
#include <linux/spi/spidev.h>
#include <sys/time.h>
#include "mms101_spi.h"
#include "mms101.h"

//==============================================================================
//{### FUNCTIONS
//==============================================================================
/*!
//==============================================================================
	@brief		Main

	@param		int argc
	@param		int **argv
	@return		int
	
//==============================================================================
	@par		Operations Description:

//==============================================================================
*/
int main(int argc, char **argv)
{
	int measNumMax;
	int measCnt;
	int drdyTime;
	struct timeval tv;

	if(argc != 2){
		mms101_usage();

		return(1);
	}

	measNumMax = (int)strtoul(argv[1], NULL, 0);
	if(measNumMax >= MEAS_NUM_MAX){
		mms101_usage();

		return(1);
	}

	// ----------------------------------------
	// Initialize
	// ----------------------------------------
	spi_init(&spi_dev);

	// ----------------------------------------
	// BOOTLOAD
	// ----------------------------------------
	if(mms101_bootload() != 0){
		return(1);
	}

	// ----------------------------------------
	// START
	// ----------------------------------------
	if(mms101_start() != 0){
		return(1);
	}

	// ----------------------------------------
	// wait for 1st conversion
	// ----------------------------------------
	usleep(WAIT_US_1ST_CONV);

	// ----------------------------------------
	// Initialize Timer & Counter
	// ----------------------------------------
	gettimeofday(&startTime, NULL);
	preGetTime.tv_sec = startTime.tv_sec;
	preGetTime.tv_usec = startTime.tv_usec;
	measureRestartLimit = RESTART_INTERVAL_NUM;
	measureCount = 0;

	// ----------------------------------------
	// Measure
	// ----------------------------------------
	printf("time[s],Fx,Fy,Fz,Mx,My,Mz\n");
	for(measCnt=0; measCnt<measNumMax; measCnt++){

		//--------------------------------------------------------------------------
		// Waiting for correction conversion completion
		//--------------------------------------------------------------------------
		// Issue Result0 command after waiting for the specified time.
		do{
			usleep(1);
			drdyTime = getElapsedTimeUs(&preGetTime, &tv);
		}while(drdyTime < INTERVAL_TIME_US);
		preGetTime.tv_sec = tv.tv_sec;
		preGetTime.tv_usec = tv.tv_usec;


		// ADC data acquisition
		if(mms101_read_adc() != 0){
			return(1);
		}
		
		// Matrix operatoin
		mms101_calc_matrix();
		
		// Check if a restart is executed
		if(measureRestartLimit != 0){
			measureCount += 1;
			if(measureCount >= measureRestartLimit){
				measureCount = 0;

				// *** Execute restart (Reacquire temperature data) ********
				if(mms101_restart() != 0){
					return(1);
				}
			}
		}
	}

	// ----------------------------------------
	// Close SPI
	// ----------------------------------------
	spi_close(&spi_dev);

	return(0);
}


/*!
//==============================================================================
	@brief		BOOTLOAD

	@param		void
	@return		int (0:OK/1:NG)
	
//==============================================================================
	@par		Operations Description:

//==============================================================================
*/
int mms101_bootload(void)
{
	uint32_t coeffTmp;
	uint8_t devNo;
	uint8_t coefNo;
	uint8_t address;

	for(devNo = 0; devNo < AFE_NUM_MAX; devNo++){

		// issue IDLE command
		tx_buf[0] = MMS101_CMD_IDLE;
		if(spi_send_data(devNo, &spi_dev, tx_buf, 1) < 0){
			printf("devNo:%d, ERROR: IDLE command\n",devNo);
			return(1); 	// error
		}

		// wait 10msec
		usleep(10000);
		
		/*
		// Debug Read Status
		tx_buf[0] = MMS101_CMD_STATUS;
		if(spi_send_after_receive_data(devNo, &spi_dev, tx_buf, 1, rx_buf, 1) < 0){
			printf("Error: read status\n");
		}else{
			printf("Idle?:%02x\n", rx_buf[0]);
		}
		*/
	
		// Read matrix calculation coefficient
		for(coefNo=0; coefNo<MMS101_COEFF_NUM_MAX; coefNo++){
			coeffTmp = 0;
			for(address=0; address<3; address++){

				// issue READ NVM command
				tx_buf[0] = MMS101_CMD_READ_NVM;													// command
				tx_buf[1] = nvm_address_map[coefNo] + address;										// address
				//printf("Matrix: %02x %02x\n", tx_buf[0], tx_buf[1]);
				if(spi_send_after_receive_data(devNo, &spi_dev, tx_buf, 2, rx_buf, 1) == 0){
					coeffTmp += (((uint32_t)rx_buf[0]) << ((2-address)*8));
				}else{
					printf("devNo:%d, ERROR: NVM(%d) read\n", devNo, tx_buf[1]);
					return(1); 	// error
				}

				// Read register(0x54)(Check if the previously read NVM data was read normally)
				tx_buf[0] = MMS101_CMD_READ_REG;		// command
				tx_buf[1] = MMS101_REG_BANK_CHK;		// address
				if(spi_send_after_receive_data(devNo, &spi_dev, tx_buf, 2, rx_buf, 1) == 0){
					if(rx_buf[0] == 0x07){
						printf("devNo:%d, ERROR: NVM(%d)\n", devNo, nvm_address_map[coefNo] + address);
						return(1); 	// error
					}
				}else{
					printf("devNo:%d, ERROR: reg(0x54) read\n", devNo);
					return(1); 		// error
				}
			}
			if(coeffTmp >= 0x00800000){
				coeffTmp |= 0xff000000;
			}
			mms101_data[devNo].coeff[coefNo] = (int32_t)coeffTmp;
			//printf("devNo:%d, coeff(%d):%x\n", devNo, coeffCnt, mms101_data[devNo].coeff[coeffCnt]);
		}
		
		// issue RESET command
		tx_buf[0] = MMS101_CMD_RESET;
		if(spi_send_data(devNo, &spi_dev, tx_buf, 1) < 0){
			printf("devNo:%d, ERROR: RESET command\n",devNo);
			return(1); 	// error
		}
	}
	
	//wait 10msec
	usleep(10000);
	
	return(0);
}


/*!
//==============================================================================
	@brief		START

	@param		void
	@return		int (0:OK/1:NG)
	
//==============================================================================
	@par		Operations Description:

//==============================================================================
*/
int mms101_start(void)
{
	uint8_t devNo;

	for(devNo = 0; devNo < AFE_NUM_MAX; devNo++){
		tx_buf[0] = MMS101_CMD_ACTIVE0; 	// Active command
		if(spi_send_data(devNo, &spi_dev, tx_buf, 1) < 0){
			printf("devNo:%d, ERROR: ACTIVE command\n", devNo);
			return(1); 	// error
		}
	}

	/*
	// Debug Read Status
	for(devNo = 0; devNo < AFE_NUM_MAX; devNo++){
		tx_buf[0] = MMS101_CMD_STATUS;
		if(spi_send_after_receive_data(devNo, &spi_dev, tx_buf, 1, rx_buf, 1) < 0){
			printf("devNo:%d, ERROR: STATUS command\n");
		}else{
			printf("devNo:%d, Active?:%02x\n", devNo, rx_buf[0]);
		}
	}
	*/

	return(0);
}


/*!
//==============================================================================
	@brief		Read ADC

	@param		void
	@return		int
//==============================================================================
	@par		Operations Description:
	
//==============================================================================
*/
int mms101_read_adc(void)
{
	uint32_t dataWork;
	uint8_t  devNo;

	memset(tx_buf, 0, SPI_DATA_SIZE);	// <-$BI,MW!)(B
	memset(rx_buf, 0, SPI_DATA_SIZE);	// <-$BI,MW!)(B

	//--------------------------------------------------------------------------
	// issue RESULT0 Command
	//--------------------------------------------------------------------------
	for(devNo=0; devNo<AFE_NUM_MAX; devNo++){
		tx_buf[0] = MMS101_CMD_RESULT0;
		if(spi_send_after_receive_data(devNo, &spi_dev, tx_buf, 1, rx_buf, 3) == 0){
			dataWork  = ((uint32_t)rx_buf[0]) << 16;
			dataWork += ((uint32_t)rx_buf[1]) << 8;
			dataWork +=  (uint32_t)rx_buf[2];
			if(dataWork >= 0x00800000){
				dataWork |= 0xff000000;
			}

			// Store in variables for matrix calculation
			mms101_data[devNo].adcData = (int32_t)dataWork;
		}else{
			printf("devNo:%d, ERROR: RESULT command\n",devNo);
			return(1); 	// error
		}
	}
	return (0);
}


/*!
//==============================================================================
	@brief		Matrix calculation

	@param		void
	@return		void
//==============================================================================
	@par		Operations Description:
	
//==============================================================================
*/
void mms101_calc_matrix(void)
{
	int64_t temp64;
	uint8_t devNo;

	for(devNo=0; devNo<AFE_NUM_MAX; devNo++)
	{
		temp64  = (int64_t)mms101_data[devNo].coeff[0] * (int64_t)mms101_data[0].adcData;
		temp64 += (int64_t)mms101_data[devNo].coeff[1] * (int64_t)mms101_data[1].adcData;
		temp64 += (int64_t)mms101_data[devNo].coeff[2] * (int64_t)mms101_data[2].adcData;
		temp64 += (int64_t)mms101_data[devNo].coeff[3] * (int64_t)mms101_data[3].adcData;
		temp64 += (int64_t)mms101_data[devNo].coeff[4] * (int64_t)mms101_data[4].adcData;
		temp64 += (int64_t)mms101_data[devNo].coeff[5] * (int64_t)mms101_data[5].adcData;
		temp64 = temp64 / (int64_t)RIGHT_BIT_SHIFT_11;

		if(temp64 < (int64_t)0xFFFFFFFFFF800000){
			mms101_data[devNo].output = (int32_t)0xFF800000;		// minimum
		}else if(temp64 > (int64_t)0x00000000007FFFFF){
			mms101_data[devNo].output = (int32_t)0x007FFFFF;		// maximum
		}else{	
			mms101_data[devNo].output = (int32_t)temp64;
		}
	}
	
	// Output matrix data
	printf("%f", getElapsedTime(&startTime, 0));
	printf(",%.3f", (double)mms101_data[DEV_FX].output / 1000);
	printf(",%.3f", (double)mms101_data[DEV_FY].output / 1000);
	printf(",%.3f", (double)mms101_data[DEV_FZ].output / 1000);
	printf(",%.5f", (double)mms101_data[DEV_MX].output / 100000);
	printf(",%.5f", (double)mms101_data[DEV_MY].output / 100000);
	printf(",%.5f", (double)mms101_data[DEV_MZ].output / 100000);
	printf("\n");

}


/*!
//==============================================================================
	@brief		RESTART

	@param		void
	@return		int (0:OK/1:NG)
	
//==============================================================================
	@par		Operations Description:

//==============================================================================
*/
int mms101_restart(void)
{
	uint8_t devNo;

	// *** Execute restart command(Reacquire temperature data) ********
	for(devNo=0; devNo<AFE_NUM_MAX; devNo++){
		tx_buf[0] = MMS101_CMD_WRITE_REG;										// 
		tx_buf[1] = 0x3F;														// 
		tx_buf[2] = 0x01;														// 
		if(spi_send_data(devNo, &spi_dev, tx_buf, 3) < 0){
			printf("devNo:%d, ERROR: RESTART command\n",devNo);
			return(1);		// error
		}
	}
	return(0);
}


/*!
//==============================================================================
	@brief		Usage

	@param		void
	@return		void
	
//==============================================================================
	@par		Operations Description:

//==============================================================================
*/
void mms101_usage(void)
{
	printf("\n");
	printf("   Usage: mms101 [measurement number]\n");
	printf("\n");
}

