//******************************************************************************
/*!
	@file		mms101_spi.c
	@brief		SPI Module Function Source File

				copyright: MITSUMI Electric Co.,LTD
	@attention	none
	@warning	MITSUMI CONFIDENTIAL
*/
//******************************************************************************
/*******************************************************************************
* History : DD.MM.YYYY Version  Description
*         : 07.10.2021 1.0.0.0  First Release
*******************************************************************************/
#define __MMS101_SPI_C__

//==============================================================================
//{### MODULES USED
//==============================================================================
#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <fcntl.h>
#include <unistd.h>
#include <string.h>
#include <linux/spi/spidev.h>
#include <sys/time.h>
#include "mms101_spi.h"
#include "mms101.h"

//==============================================================================
//{### FUNCTIONS
//==============================================================================
/*!
//==============================================================================
	@brief		SPI Initialize

	@param		spi_device_t *spi_device
	@return		int
	
//==============================================================================
	@par		Operations Description:

//==============================================================================
*/
int spi_init(spi_device_t *spi_device)
{
	int status = 0;

	memset(spi_device, 0, sizeof(spi_device_t));

	spi_csb_init(spi_device);

	spi_device->fd = open(DEV_SPI_PATH, O_RDWR);
	if(spi_device->fd < 0){
		printf("Error: open spi device\n");
		status = -1;
	}

	spi_device->mode	= SPI_MODE;
	spi_device->tr[0].speed_hz		= SPI_SPEED_HZ;
	spi_device->tr[0].delay_usecs	= SPI_DELAY_USECS;
	spi_device->tr[0].bits_per_word	= SPI_SPI_BITS;
	spi_device->tr[0].cs_change		= SPI_CS_CHANGE;

	if(ioctl(spi_device->fd, SPI_IOC_WR_MODE, &(spi_device->mode)) < 0){
		printf("Error: set spi mode(wr)\n");
		status = -1;
	}

	if(ioctl(spi_device->fd, SPI_IOC_RD_MODE, &(spi_device->mode)) < 0){
		printf("Error: set spi mode(rd)\n");
		status = -1;
	}

	if(ioctl(spi_device->fd, SPI_IOC_WR_MAX_SPEED_HZ, &(spi_device->tr[0].speed_hz)) < 0){
		printf("Error: set spi speed(wr)\n");
		status = -1;
	}

	if(ioctl(spi_device->fd, SPI_IOC_RD_MAX_SPEED_HZ, &(spi_device->tr[0].speed_hz)) < 0){
		printf("Error: set spi speed(rd)\n");
		status = -1;
	}

	return status;
}

/*!
//==============================================================================
	@brief		SPI Chip Select Initialize

	@param		spi_device_t *spi_device
	@return		int
	
//==============================================================================
	@par		Operations Description:

//==============================================================================
*/
int spi_csb_init(spi_device_t *spi_device)
{
	int status = 0;
	uint8_t devNo;
	int fd = 0;
	char devName[50];
	char devNameBase1[50] = DEV_GPIO_PATH1;
	char devNameBase2[50] = DEV_GPIO_PATH2;

	for(devNo=0; devNo<AFE_NUM_MAX; devNo++){
		// setting reserved
		strncpy(devName, devNameBase1, 40);
		strncat(devName, "/export", 40);
		fd = open(devName, O_WRONLY);
		if(fd < 0){
			printf("GPIO reserve error.\n");
			status = -1;
		}
		write(fd, csGpioNo[devNo], 2);
		close(fd);

		// wait 0.5s
		usleep(50 * 1000);

		// setting output
		strncpy(devName, devNameBase2, 40);
		strncat(devName, csGpioNo[devNo], 40);
		strncat(devName, "/direction", 40);
		fd = open(devName, O_WRONLY);
		if(fd < 0){
			printf("GPIO direction error.\n");
			status = -1;
		}
		write(fd, "out", 4);
		close(fd);
	}

	spi_csb_open(spi_device);

	for(devNo=0; devNo<AFE_NUM_MAX; devNo++){
		if(spi_disable(devNo, spi_device) < 0){
			status = -1;
		}
	}

	return status;
}

/*!
//==============================================================================
	@brief		Open SPI Chip Select Port(GPIO)

	@param		spi_device_t *spi_device
	@return		int
	
//==============================================================================
	@par		Operations Description:

//==============================================================================
*/
int spi_csb_open(spi_device_t *spi_device)
{
	char devNameBase[50] = DEV_GPIO_PATH2;
	char devName[50];
	uint8_t devNo;
	int status = 0;

	for(devNo=0; devNo<AFE_NUM_MAX; devNo++){
		strncpy(devName, devNameBase, 40);
		strncat(devName, csGpioNo[devNo], 40);
		strncat(devName, "/value", 40);
		spi_device->fd_csb[devNo] = open(devName, O_WRONLY);
		if(spi_device->fd_csb[devNo] < 0){
			status = -1;
		}
	}

	return status;
}

/*!
//==============================================================================
	@brief		Close SPI Chip Select Port(GPIO)

	@param		spi_device_t *spi_device
	@return		int
	
//==============================================================================
	@par		Operations Description:

//==============================================================================
*/
int spi_csb_close(spi_device_t *spi_device)
{
	uint8_t devNo;
	int status = 0;

	for(devNo=0; devNo<AFE_NUM_MAX; devNo++){
		if(close(spi_device->fd_csb[devNo]) < 0){
			status = -1;
		}
	}

	return status;
}

/*!
//==============================================================================
	@brief		SPI Enable Module

	@param		uint8_t devNo
	@param		spi_device_t *spi_device
	@return		int
	
//==============================================================================
	@par		Operations Description:

//==============================================================================
*/
int spi_enable(uint8_t devNo, spi_device_t *spi_device)
{
	return write(spi_device->fd_csb[devNo], "0", 2);
}

/*!
//==============================================================================
	@brief		SPI Disable Module

	@param		uint8_t devNo
	@param		spi_device_t *spi_device
	@return		int
	
//==============================================================================
	@par		Operations Description:

//==============================================================================
*/
int spi_disable(uint8_t devNo, spi_device_t *spi_device)
{
	return write(spi_device->fd_csb[devNo], "1", 2);
}

/*!
//==============================================================================
	@brief		SPI Close

	@param		spi_device_t *spi_device
	@return		int
	
//==============================================================================
	@par		Operations Description:

//==============================================================================
*/
int spi_close(spi_device_t *spi_device)
{
	int status = 0;

	if(spi_csb_close(spi_device) < 0){
		status = -1;
	}

	if(close(spi_device->fd) < 0){
		status = -1;
	}

	return status;
}

/*!
//==============================================================================
	@brief		SPI Send Data Module

	@param		uint8_t devNo  : device number
	@param		spi_device_t *spi_device  : spi device infomation
	@param		uint8_t *txbuf  : transmitting data buffer
	@param		uint8_t txlen   : number of transmitting data
	@return		int
	
//==============================================================================
	@par		Operations Description:

//==============================================================================
*/
int spi_send_data(uint8_t devNo, spi_device_t *spi_device, uint8_t *txbuf, uint8_t txlen)
{
	int status = 0;

	if(spi_enable(devNo, spi_device) < 0){
		status = -1;
	}

	spi_device->tr[0].tx_buf = (ulong)txbuf;
	spi_device->tr[0].rx_buf = 0;
	spi_device->tr[0].len = txlen;

	if(ioctl(spi_device->fd, SPI_IOC_MESSAGE(1), spi_device->tr) < 0){
		status = -1;
	}

	if(spi_disable(devNo, spi_device) < 0){
		status = -1;
	}

	return status;
}

/*!
//==============================================================================
	@brief		SPI Receive Data Module

	@param		uint8_t devNo  : device number
	@param		spi_device_t *spi_device  : spi device infomation
	@param		uint8_t *rxbuf  : received data buffer
	@param		uint8_t rxlen   : number of received data
	@return		int
	
//==============================================================================
	@par		Operations Description:

//==============================================================================
*/
int spi_receive_data(uint8_t devNo, spi_device_t *spi_device, uint8_t *rxbuf, uint8_t rxlen)
{
	int status = 0;

	if(spi_enable(devNo, spi_device) < 0){
		status = -1;
	}

	spi_device->tr[0].tx_buf = 0;
	spi_device->tr[0].rx_buf = (ulong)rxbuf;
	spi_device->tr[0].len = rxlen;

	if(ioctl(spi_device->fd, SPI_IOC_MESSAGE(1), spi_device->tr) < 0){
		status = -1;
	}

	if(spi_disable(devNo, spi_device) < 0){
		status = -1;
	}

	return status;
}

/*!
//==============================================================================
	@brief		SPI Send after Receive Data Module

	@param		uint8_t devNo  : number of device
	@param		spi_device_t *spi_device  : spi device infomation
	@param		uint8_t *txbuf  : transmitting data buffer
	@param		uint8_t txlen   : number of transmitting data
	@param		uint8_t *rxbuf  : received data buffer
	@param		uint8_t rxlen   : number of received data
	@return		int
	
//==============================================================================
	@par		Operations Description:

//==============================================================================
*/
int spi_send_after_receive_data(uint8_t devNo, spi_device_t *spi_device, uint8_t *txbuf, uint8_t txlen, uint8_t *rxbuf, uint8_t rxlen)
{
	int status = 0;

	if(spi_enable(devNo, spi_device) < 0){
		status = -1;
	}

	// transmit
	spi_device->tr[0].tx_buf = (ulong)txbuf;
	spi_device->tr[0].rx_buf = 0;
	spi_device->tr[0].len = txlen;
	if(ioctl(spi_device->fd, SPI_IOC_MESSAGE(1), spi_device->tr) < 0){
		status = -1;
	}

	if(txbuf[0] == MMS101_CMD_READ_NVM){
		if(spi_check_busy(spi_device, 1000) != 0){
			status = -1;	// timeout: 1000us
		}
	}

	// receive
	spi_device->tr[0].tx_buf = 0;
	spi_device->tr[0].rx_buf = (ulong)rxbuf;
	spi_device->tr[0].len = rxlen;
	if(ioctl(spi_device->fd, SPI_IOC_MESSAGE(1), spi_device->tr) < 0){
		status = -1;
	}

	if(spi_disable(devNo, spi_device) < 0){
		status = -1;
	}

	return status;
}

/*!
//==============================================================================
	@brief 		SPI BUSY CHECK

	@param		spi_device_t *spi_device  : spi device infomation
	@param 		long timeout(usec)
	@return 	int

//==============================================================================
	@par 		Operations Description:

//==============================================================================
*/
int spi_check_busy(spi_device_t *spi_device, long timeout)
{
	int status = 0;
	uint8_t buf[SPI_DATA_SIZE];
	struct timeval tv;

	gettimeofday(&tv, NULL);
	do{
		// receive
		spi_device->tr[0].tx_buf = 0;
		spi_device->tr[0].rx_buf = (ulong)buf;
		spi_device->tr[0].len = 1;
		if(ioctl(spi_device->fd, SPI_IOC_MESSAGE(1), spi_device->tr) < 0){
			status = -1;
			break;
		}
		
		if(getElapsedTimeUs(&tv, NULL) > timeout){
			status = -1;
			break;
		}
	}while(buf[0] != 0x01);

	return status;
}

/*!
//==============================================================================
	@brief 		Get Elapsed Time(usec)

	@param		struct timeval *tvStart : Base Time
	@param		struct timeval *tvUpdate : Copy Now Time (Null:Not copy)
	@return 	int : elapsed time(usec) (<= 2147483647usec)

//==============================================================================
	@par 		Operations Description:

//==============================================================================
*/
int getElapsedTimeUs(struct timeval *tvStart, struct timeval *tvUpdate){
	struct timeval tvNow;
	int elapsed;

	gettimeofday(&tvNow, NULL);

	elapsed = (tvNow.tv_sec - tvStart->tv_sec) * 1000000;
	elapsed += (tvNow.tv_usec - tvStart->tv_usec);

	if(tvUpdate != NULL){
		tvUpdate->tv_sec = tvNow.tv_sec;
		tvUpdate->tv_usec = tvNow.tv_usec;
	}

	return elapsed;
}

/*!
//==============================================================================
	@brief 		Read Elapsed Time(sec)

	@param		struct timeval *tvStart : Base Time
	@param		struct timeval *tvUpdate : Copy Now Time (Null:Not copy)
	@return 	double : elapsed time(sec)

//==============================================================================
	@par 		Operations Description:

//==============================================================================
*/
double getElapsedTime(struct timeval *tvStart, struct timeval *tvUpdate){
	struct timeval tvNow;
	double elapsed;

	gettimeofday(&tvNow, NULL);

	elapsed = (double)tvNow.tv_sec + ((double)tvNow.tv_usec) / 1000000;
	elapsed -= ((double)tvStart->tv_sec + ((double)tvStart->tv_usec) / 1000000);

	if(tvUpdate != NULL){
		tvUpdate->tv_sec = tvNow.tv_sec;
		tvUpdate->tv_usec = tvNow.tv_usec;
	}

	return elapsed;
}

